/*-
 * Copyright (C)2018..2022 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2018..2022 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: db_mssql.c,v 1.39 2022/01/08 22:59:07 babolo Exp $"

#define BLIN_COMPAT      4
#define Bpars_COMPAT     3
#define MULAR_COMPAT     0
#define MIFE_COMPAT      4
#define RECOBE_COMPAT    4
#define PGOBLIN_COMPAT   4
#define PGOBLIN_INTERNAL 1
#define PGOBLIN_MODULE   pgoblin_module

#include <sys/limits.h>
#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sybfront.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdio.h>
#include <errno.h>
#include <sybdb.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include <multilar.h>
#include <mife.h>
#include "pgoblin.h"

#define INDX(R, C)  (R * rio->pq->ncol + C)
#define MULAR_FLAGS (MULAR_CHAR | MULAR_ZERO | MULAR_PFRE | MULAR_STRI | MULAR_UPPE)
#define TYPEMAX 192

static char *types[TYPEMAX] =
{ ""         , ""        , ""           , ""                 /* 00 */
, ""         , ""        , ""           , ""                 /* 04 */
, ""         , ""        , ""           , ""                 /* 08 */
, ""         , ""        , ""           , ""                 /* 0C */
, ""         , ""        , ""           , ""                 /* 10 */
, ""         , ""        , ""           , ""                 /* 14 */
, ""         , ""        , ""           , ""                 /* 18 */
, ""         , ""        , ""           , "VOID"             /* 1C */
, ""         , ""        , "IMAGE"      , "TEXT"             /* 20 */
, "VARBINARY", ""        , "INTN"       , "VARCHAR"          /* 24 */
, "MSDATE"   , "MSTIME"  , "MSDATETIME2", "MSDATETIMEOFFSET" /* 28 */
, ""         , "BINARY"  , ""           , "CHAR"             /* 2C */
, "INT1"     , "DATE"    , "BIT"        , "TIME"             /* 30 */
, "INT2"     , ""        , ""           , ""                 /* 34 */
, "INT4"     , ""        , "DATETIME4"  , "REAL"             /* 38 */
, "MONEY"    , "DATETIME", "FLT8"       , ""                 /* 3C */
, ""         , ""        , ""           , ""                 /* 40 */
, ""         , ""        , ""           , ""                 /* 44 */
, ""         , ""        , ""           , ""                 /* 48 */
, ""         , ""        , ""           , ""                 /* 4C */
, ""         , ""        , ""           , ""                 /* 50 */
, ""         , ""        , ""           , ""                 /* 54 */
, ""         , ""        , ""           , ""                 /* 58 */
, ""         , ""        , ""           , ""                 /* 5C */
, ""         , ""        , ""           , "NTEXT"            /* 60 */
, ""         , ""        , ""           , "NVARCHAR"         /* 64 */
, "BITN"     , ""        , "DECIMAL"    , ""                 /* 68 */
, "NUMERIC"  , "FLTN"    , "MONEYN"     , "DATETIMN"         /* 6C */
, ""         , ""        , ""           , ""                 /* 70 */
, ""         , ""        , ""           , ""                 /* 74 */
, ""         , ""        , "MONEY4"     , ""                 /* 78 */
, ""         , ""        , ""           , "INT8"             /* 7C */
, ""         , ""        , ""           , ""                 /* 80 */
, ""         , ""        , ""           , ""                 /* 84 */
, ""         , ""        , ""           , ""                 /* 88 */
, ""         , ""        , ""           , ""                 /* 8C */
, ""         , ""        , ""           , ""                 /* 90 */
, ""         , ""        , ""           , ""                 /* 94 */
, ""         , ""        , ""           , ""                 /* 98 */
, ""         , ""        , ""           , ""                 /* 9C */
, ""         , ""        , ""           , ""                 /* A0 */
, ""         , ""        , ""           , ""                 /* A4 */
, ""         , ""        , ""           , ""                 /* A8 */
, ""         , ""        , ""           , ""                 /* AC */
, ""         , ""        , ""           , ""                 /* B0 */
, ""         , ""        , ""           , ""                 /* B4 */
, ""         , ""        , ""           , "BIGDATETIME"      /* B8 */
, "BIGTIME"  , ""        , ""           , ""                 /* BC */
};

struct pgoblin_realpq {
    BLIN_flag         flags;
/*                    PGOBLIN_DB_TYPE                                                                */
#   define            PGOBLIN_TINIT   0x010000
#   define            PGOBLIN_BUFF    0x040000
    int               ncol;
    mular_descriptor *md;
    char             *mess;
    int              *status;
    DBCOL            *colinfo;
    char            **buffer;
    u_char           *reco;
    int               nrow;
};

typedef struct pgoblin_db_conn {
    BLIN_flag        flags;
/*                   PGOBLIN_DB_TYPE                                                                 */
#   define           PGOBLIN_STREAMI 0x010000
#   define           PGOBLIN_STREAMT 0x020000
#   define           PGOBLIN_STREAME 0x040000
#   define           PGOBLIN_ALLOUT  (PGOBLIN_STREAMI | PGOBLIN_STREAMT | PGOBLIN_STREAME)
#   define           PGOBLIN_COPYIN  0x100000
#   define           PGOBLIN_COPYST  0x200000
#   define           PGOBLIN_COPYEN  0x400000
#   define           PGOBLIN_ALLIN   (PGOBLIN_COPYIN | PGOBLIN_COPYST | PGOBLIN_COPYEN)
    pgoblin_realpq  *res;
    LOGINREC        *login;
    DBPROCESS       *dbproc;
} pgoblin_db_conn;

static size_t  mdp[] = {1024, 512, 512};
pgoblin_dbases PGOBLIN_MODULE;

static pgoblin_realpq *
/**********************************************************************
 **                                                                  **/
resv(pgoblin_rdb *rdb) {                                            /**
 **                                                                  **
 **********************************************************************/
    pgoblin_realpq *rsv = NULL;

#   define blin_internal_flags (rdb->flags & BLIN_MASK)
    ifBLIN_QX4("+ %08X", rdb->flags);
    if  (!(rsv = calloc(1, sizeof(pgoblin_realpq)))) {
        ifBLIN_QW0("No mem");
        goto out;
    } 
    rsv->flags = rdb->flags & (BLIN_MASK | PGOBLIN_DB_TYPE);
    rsv->reco = recobe_collection.domains[1]->reconv(recobe_collection.domains[1]->convbyname("wk"));
out:
    ifBLIN_QX4("- %"BLIN_X, BLIN_I(rsv));
    return(rsv);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
collect(DBPROCESS *dbproc, pgoblin_realpq *res, int tupl) {         /**
 **                                                                  **
 **********************************************************************/
    RETCODE           erc;
    int               ex      = EX_OK;

    if  (!res) goto out;
#   define blin_internal_flags (res->flags & BLIN_MASK)
    if  (!!(res->flags & PGOBLIN_TINIT)) goto out;
    res->ncol = dbnumcols(dbproc);
    res->nrow = 0;
    res->flags |= PGOBLIN_TINIT;
    if  (!tupl || (0 >= res->ncol)) goto out;
    if  (!(res->colinfo = malloc(res->ncol * sizeof(DBCOL)))) {
        ifBLIN_QX0("No mem %"BLIN_D, res->ncol * sizeof(DBCOL));
        ERROUT(-EX_SOFTWARE, EINVAL);
    }
    for (int ncol = 0; ncol < res->ncol; ++ncol) {
        if  (SUCCEED != dbcolinfo(dbproc, CI_REGULAR, ncol + 1, 0, &res->colinfo[ncol])) {
            ifBLIN_QX0("dbcolinfo[%d (+1)]", ncol);
            ERROUT(-EX_SOFTWARE, EPROTONOSUPPORT);
        }
        ifBLIN_QX5( "%d (%d)[%d(%d)]=%s~"
                  , ncol
                  , res->colinfo[ncol].Type
                  , res->colinfo[ncol].MaxLength
                  , res->colinfo[ncol].Name
                  );
    }
    if  (!(res->status = calloc(res->ncol, sizeof(int)))) {
        ifBLIN_QX0("no RAM %d x %d", (int)res->ncol, (int)sizeof(int));
        ERROUT(-EX_OSERR, ENOMEM);
    }
    if  (!(res->buffer = calloc(res->ncol, sizeof(char*)))) {
        ifBLIN_QX0("no RAM %d x %d", (int)res->ncol, (int)sizeof(char*));
        ERROUT(-EX_OSERR, ENOMEM);
    }
    for (int ncol = 0; ncol < res->ncol; ncol++) {
        int sz = dbprcollen(dbproc, ncol + 1);

        if  (65535 < sz) sz = 65535;
        ifBLIN_QX6("==[%d]%"BLIN_X"(%d)", ncol, BLIN_I(res->buffer[ncol]), sz);
        if  (!(res->buffer[ncol] = calloc(1, sz + 1))) {
            ifBLIN_QX0("no RAM %d", sz + 1);
            ERROUT(-EX_OSERR, ENOMEM);
        }
        erc = dbbind(dbproc, ncol + 1, NTBSTRINGBIND, sz + 1, (BYTE*)res->buffer[ncol]);
        if  (FAIL == erc) {
            ifBLIN_QX0("dbbind");
            ERROUT(-EX_SOFTWARE, EINVAL);
        }
        ifBLIN_QX6("<>[%d]%"BLIN_X, ncol, &res->status[ncol]);
        res->status[ncol] = 0;
        if  (FAIL == (erc = dbnullbind(dbproc, ncol + 1, &res->status[ncol]))) {
            ifBLIN_QX0("dbbind");
            ERROUT(-EX_SOFTWARE, EINVAL);
    }   }
out:;
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
msg_handler( DBPROCESS *dbproc                                      /**/
/**/       , DBINT      msgno                                       /**/
/**/       , int        msgstate                                    /**/
/**/       , int        severity                                    /**/
/**/       , char      *msgtext                                     /**/
/**/       , char      *srvname                                     /**/
/**/       , char      *procname                                    /**/
/**/       , int        line                                        /**/
/**/       ) {                                                      /**
 **                                                                  **
 **********************************************************************/
    enum {changed_database = 5701, changed_language = 5703 };

    if  (0 < msgno) {
        ifBLIN_QX1("Msg %ld, Level %d, State %d", (long)msgno, severity, msgstate);
        if  (0 < strlen(srvname)) ifBLIN_QX0("Server '%s', ", srvname);
        if  (0 < strlen(procname)) ifBLIN_QX0("Procedure '%s', ", procname);
        if  (0 < line) ifBLIN_QX0("Line %d", line);
    }
    if  (10 < severity) {
        ifBLIN_QX0("%s: severity %d > 10", msgtext, severity);
    } else {
        ifBLIN_QX1("%d: %s", severity, msgtext);
    }
    return(0);
}

static int
/**********************************************************************
 **                                                                  **/
err_handler( DBPROCESS *dbproc                                      /**/
/**/       , int        severity                                    /**/
/**/       , int        dberr                                       /**/
/**/       , int        oserr                                       /**/
/**/       , char      *dberrstr                                    /**/
/**/       , char      *oserrstr                                    /**/
/**/       ) {	                                                    /**
 **                                                                  **
 **********************************************************************/
    if  (!!dberr) {
        ifBLIN_QX0("Msg %d, Level %d=%s", dberr, severity, dberrstr);
    } else {
        ifBLIN_QX0("DB-LIBRARY error: %s", dberrstr);
    }
    return(INT_CANCEL);
}

static int
/**********************************************************************
 **                                                                  **/
clear(pgoblin_exenv *exenv, pgoblin_nr nio) {                       /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    pgoblin_rio    *rio;
    pgoblin_realpq *rsv;
    int             ex    = EX_OK;

    GET_RIO(rio, exenv->options, nio);
    if  (!!rio && !!rio->pq) {
        rsv = rio->pq;
        ifBLIN_QX4("+ %"BLIN_X" %"BLIN_X, BLIN_I(rio), BLIN_I(rsv));
        if  (!!rsv->md) mular_destroy(rsv->md);
        if  (!!rsv->colinfo) free(rsv->colinfo);
        if  (!!rsv->buffer) {
            for (int ncol = 0; ncol < rsv->ncol; ++ncol) {
                if  (!!rsv->buffer[ncol]) free(rsv->buffer[ncol]);
            }
            free(rsv->buffer);
        }
        if  (!!rsv->status) free(rsv->status);
        if  (!!rsv->reco) free(rsv->reco);
        free(rio->pq);
        rio->pq = NULL;
    }
    ifBLIN_QX4("- %d", ex);
out:
    return(ex);
#   undef blin_internal_flags
}

static ssize_t
/******************************************************************************
 **                                                                          **/
valinfo(pgoblin_exenv *exenv, pgoblin_nr nio, int kind, int row, int col) { /**
 **                                                                          **
 ******************************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    pgoblin_rio  *rio;
    ssize_t       ex = PGOBLIN_SSIZE_MIN;

    GET_RIO(rio, exenv->options, nio);
    switch (kind) {
    case PGOBLIN_IsNull:
        if  (rio->pq) {
            if  ((row < 0) || (row >= rio->pq->nrow) || (col < 0) || (col >= rio->pq->ncol)) {
                ifBLIN_QX1( "+ row %d %s in [0..%d], col %d %s in [0..%d]"
                          , row, (row >= 0 && row < rio->pq->nrow) ? "" : "NOT", rio->pq->nrow
                          , col, (col >= 0 && col < rio->pq->ncol) ? "" : "NOT", rio->pq->ncol
                          );
                errno = EDOOFUS;
            } else if (!!(rio->pq->flags & PGOBLIN_BUFF)) {
                if  (0 < rio->pq->status[INDX(row, col)]) {
                    errno = ENOMEM;
                } else {
                    ex = !!rio->pq->status[INDX(row, col)];
                }
            } else if (!!rio->pq->md) {
                ex = !*(char**)mular_getix(rio->pq->md, INDX(row, col));
            } else {
                ifBLIN_QX0("NOTHING instead of MULAR");
                errno = EDOOFUS;
            }
        } else if (row == 0 && col == 0) {
            ex = 1;
        } else {
            ifBLIN_QX1("row %d col %d in emty", row, col);
            errno = EDOOFUS;
        }
        break;
    case PGOBLIN_Length:
        switch (valinfo(exenv, nio, PGOBLIN_IsNull, row, col)) {
        case  1:
            ex = 0;
            break;
        case  0:
            if  (!!(rio->pq->flags & PGOBLIN_BUFF)) {
                ex = strlen(rio->pq->buffer[INDX(row, col)]);
            } else if (!!rio->pq->md) {
                ex = strlen(*(char**)mular_getix(rio->pq->md, INDX(row, col)));
            } else {
                ifBLIN_QX0("NOTHING instead of MULAR");
                errno = EDOOFUS;
            }
            break;
        default:
            errno = EDOOFUS;
        }
        break;
    case PGOBLIN_TypeOid:
        if  (!rio->pq->colinfo || !(rio->pq->ncol > col)) {
            errno = EDOOFUS;
        } else {
            ex = rio->pq->colinfo[col].Type;
        }
        break;
    case PGOBLIN_TypeMod:
        if  (!rio->pq->colinfo || !(rio->pq->ncol > col)) {
            errno = EDOOFUS;
        } else {
            ex = rio->pq->colinfo[col].Precision;
        }
        break;
    case PGOBLIN_ColSize:
        if  (!rio->pq->colinfo || !(rio->pq->ncol > col)) {
            errno = EDOOFUS;
        } else {
            ex = rio->pq->colinfo[col].MaxLength;
        }
        break;
    default:
        errno = ENOTSUP;
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}
	
static int
/**********************************************************************
 **                                                                  **/
init(pgoblin_main *options, int slot) {                             /**
 **                                                                  **
 **********************************************************************/
    int ex = EX_OK;

#   define blin_internal_flags (options->flags & BLIN_MASK)
    PGOBLIN_MODULE.flags = (PGOBLIN_MODULE.flags & ~(BLIN_MASK | PGOBLIN_DB_TYPE))
                         | (options->flags & BLIN_MASK)
                         | (slot & PGOBLIN_DB_TYPE)
    ;
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
fini(pgoblin_main *options) {                                       /**
 **                                                                  **
 **********************************************************************/
    return(0);
}

static int
/**********************************************************************
 **                                                                  **/
shurecon(pgoblin_exenv *exenv, pgoblin_nr ndb) {                    /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    pgoblin_db_conn *conn;
    pgoblin_rdb     *rdb;
    int              ex   = EX_OK;

    GET_CON(rdb, exenv->options, ndb);
    ifBLIN_QX2("+ %08X", rdb->flags);
    if  (!!rdb->conn) goto out;
    ifBLIN_QX2("= %d", (int)sizeof(pgoblin_db_conn));
    if  (!(conn = calloc(1, sizeof(pgoblin_db_conn)))) {
        ifBLIN_QX0("no MEM");
        ERROUT(EX_OSERR, ENOMEM);
    }
    conn->flags = PGOBLIN_MODULE.flags & (BLIN_MASK | PGOBLIN_DB_TYPE);
    MARK_R_CONN_GO(ndb);;
    ;;  rdb->conn = conn;
    ;;  rdb->flags = PGOBLIN_CLOSECONN | (rdb->flags & ~(BLIN_MASK | PGOBLIN_DB_TYPE)) | conn->flags;
    MARK_R_CONN_WENT(ndb);;
    if  (FAIL == dbinit()) {
	ifBLIN_QX0("dbinit");
        ERROUT(EX_SOFTWARE, EDOOFUS);
    }
    dberrhandle(err_handler);
    dbmsghandle(msg_handler);
    if  (!(conn->login = dblogin())) {
        ifBLIN_QX0("dblogin");
        ERROUT(EX_SOFTWARE, EDOOFUS);
    }
    DBSETLUSER(conn->login, rdb->username);
    DBSETLPWD(conn->login, rdb->password);
    if  (!(conn->dbproc = dbopen(conn->login, rdb->host))) {
        ifBLIN_QX0("dbopen");
        ERROUT(EX_SOFTWARE, EDOOFUS);
    }
    if  (rdb->dbname && (FAIL == dbuse(conn->dbproc, rdb->dbname))) {
        ifBLIN_QX0("dbuse");
        ERROUT(EX_SOFTWARE, EDOOFUS);
    }
out:
    ifBLIN_QX2("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/********************************************************************************
 **                                                                            **/
query(pgoblin_exenv *exenv, pgoblin_nr ndb, pgoblin_nr nou, pgoblin_nr nct) { /**
 **                                                                            **
 ********************************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    int              row_code;
    BLIN_flag        cond;
    pgoblin_db_conn *conn = NULL;
    pgoblin_rdb     *rdb;
    RETCODE          erc;
    int              idx;
    pgoblin_rio     *rou;
    pgoblin_rio     *rct;
    pgoblin_realpq **rsv  = NULL;
    int              ex   = EX_OK;
    const char      *q;
    char            *p;
    static const babolo_lexor copy = {(u_char*)
    "\000..." "...."    "  | "    "  .."    "...."    "...."    "...."    "...."
    " .."    "..."    "...."    ".-F."    "IIII"    "IIII"    "II.."    "...."
    ".EEC"    "EEEE"    "EEEE"    "EEEO"    "PEEE"    "EEEE"    "EYE."    "...E"
    ".EEC"    "EEEE"    "EEEE"    "EEEO"    "PEEE"    "EEEE"    "EYE."    "...."
    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"
    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"
    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"
    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"    "EEEE"
    , NULL, Bpars_AABS | Bpars_CEND, 0x1C, 0x1C, 1
    ,{/*00*/ 3, ' ', 0x00, '|', 0x00, '-', 0x05, 'C', 0x09, 0xFF
     ,/*05*/ 0, '-', 0x07, 0xFF
     ,/*07*/ 0, '|', 0x00, 0x07
     ,/*09*/ 0, 'O', 0x0B, 0xFF
     ,/*0B*/ 0, 'P', 0x0D, 0xFF
     ,/*0D*/ 0, 'Y', 0x0F, 0xFF
     ,/*0F*/ 8, '', 0xFE, ' ', 0x0F, '|', 0x0F, '-', 0x19, 'C', 0xFD
              , 'O', 0xFD, 'P', 0xFD, 'Y', 0xFD, 'E', 0xFD, 0xFF
     ,/*19*/ 0, '-', 0x1A, 0xFF
     ,/*1A*/ 0, '|', 0x0F, 0x1A
    }};
    /*****************
        * 0 ...
        *   ' --\n'...
        * 1 'COPY "'
        * 1 'COPY --\n "'
        * 1 "COPY '"
        * 1 'COPY\n"'
        * 2 'COPY x'
        * 2 'COPY --\n x'
    *****************/

    GET_CON(rdb, exenv->options, ndb);
    GET_RIO(rou, exenv->options, nou);
    GET_RIO(rct, exenv->options, nct);
    cond = rdb->flags & PGOBLIN_CR_MASK;
    conn = rdb->conn;
    ifBLIN_QX2("+ %d=%s~", cond, rct->text);
    if  (conn->flags & (PGOBLIN_ALLOUT | PGOBLIN_ALLIN)) {
        ifBLIN_QX0("COPY state");
        ERROUT(EX_DATAERR, ENODEV);
    }
    rsv = &rou->pq;
    *rsv = NULL;
    q = rct->text;
    switch (cond) {
    case PGOBLIN_CR_OUT:
        switch (babolo_goword(&copy, &q)) {
        case 1:
            if  (!*q) {
                ERROUT(EX_DATAERR, EINVAL);
            }
            q--;
            if  (*q == '\'') {
                for (idx = 1; ;) {
                    if  (q[idx++] == '\'' && q[idx] != '\'') break;
                    if  (!q[idx]) {
                        ERROUT(EX_DATAERR, EINVAL);
                }   }
            } else if (*q == '\"') {
                for (idx = 1; ;) {
                    if  (q[idx++] == '\"' && q[idx] != '\"') break;
                    if  (!q[idx]) {
                        ERROUT(EX_DATAERR, EINVAL);
                }   }
            } else {
                ERROUT(EX_DATAERR, EINVAL);
            }
            idx++;
            ifBLIN_QX4("goword quoted %d", idx);
            goto p3;
        case 2:
            q--;
            for (idx = 1; ;) if ((copy.retable[q[idx++] & 0xFF] & 0xC0) != 0x40) break;
            ifBLIN_QX4("goword direct %d", idx);
        p3: if  (!(p = strndup(q, --idx))) {
                ifBLIN_QX0("no mem #5");
                ERROUT(EX_OSERR, ENOMEM);
            }
            if  (asprintf((char**)&rct->text, "SELECT * FROM %s", p) < 0) {
                ifBLIN_QX0("no mem #6");
                ERROUT(EX_OSERR, ENOMEM);
            }
            ifBLIN_QX3("qry=%s~", rct->text);
        }
            /* FALLTHROUGH */
    case PGOBLIN_CR_SSEL:
        if  (FAIL == dbfcmd(conn->dbproc, rct->text)) {
            ifBLIN_QX0("dbfcmd=%s~", rct->text);
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        if  (FAIL == dbsqlexec(conn->dbproc)) {
            ifBLIN_QX0("dbsqlexec=%s~", rct->text);
            ERROUT(EX_DATAERR, EDOOFUS);
        }
        if  (NO_MORE_RESULTS != (erc = dbresults(conn->dbproc))) {
            if  (FAIL == erc) {
                ifBLIN_QX0("dbresults");
                ERROUT(EX_DATAERR, EDOOFUS);
            }
            if  (!!*rsv) clear(exenv, nou);
            if  (!(*rsv = resv(rdb))) {
                ifBLIN_QX0("no RAM");
                ERROUT(EX_OSERR, ENOMEM);
            }
            if  (0 > (collect(conn->dbproc, *rsv, 1))) {
                ifBLIN_QX0("collect");
                goto out;
        }   }
        conn->flags |= PGOBLIN_STREAMI;
        break;
    case PGOBLIN_CR_OK:
        if  (FAIL == dbfcmd(conn->dbproc, rct->text)) {
            ifBLIN_QX0("dbfcmd=%s~", rct->text);
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        if  (FAIL == dbsqlexec(conn->dbproc)) {
            ifBLIN_QX0("dbsqlexec=%s~", rct->text);
            ERROUT(EX_DATAERR, EDOOFUS);
        }
        for (int i = 0; NO_MORE_RESULTS != (erc = dbresults(conn->dbproc)); ++i) {
            if  (FAIL == erc) {
                ifBLIN_QX0("dbresults %d", i);
                ERROUT(EX_DATAERR, EDOOFUS);
            }
            if  (!!*rsv) clear(exenv, nou);
            if  (!(*rsv = resv(rdb))) {
                ifBLIN_QX0("no RAM");
                ERROUT(EX_OSERR, ENOMEM);
            }
            if  (0 > (collect(conn->dbproc, *rsv, 0))) {
                ifBLIN_QX0("collect");
                goto out;
            }
            for (int j = 0; NO_MORE_ROWS != (row_code = dbnextrow(conn->dbproc)); ++j) {
                switch (row_code) {
                case REG_ROW:
                    break;
                case BUF_FULL:
                    ifBLIN_QX0("BUF_FULL *rsv=%d row=%d", i, j);
                    break;
                case FAIL:
                    ifBLIN_QX0("FAIL *rsv=%d row=%d", i, j);
                    ERROUT(-EX_SOFTWARE, EINVAL);
                default:
                    ifBLIN_QX0("%d ignored, *rsv=%d row=%d", i, j);
                    break;
            }   }
            if  (0 > DBCOUNT(conn->dbproc)) {
                ifBLIN_QX1("%d: row: DBCOUNT %d", i, DBCOUNT(conn->dbproc));
            } else if ((*rsv)->nrow == DBCOUNT(conn->dbproc)) {
            } else if (!(*rsv)->nrow) {
                ifBLIN_QX1("%d:%d row: 0 > DBCOUNT %d", i, DBCOUNT(conn->dbproc));
            } else if ((*rsv)->nrow == DBCOUNT(conn->dbproc)) {
                ifBLIN_QX0("%d: row: %d > DBCOUNT %d", i, (*rsv)->nrow, DBCOUNT(conn->dbproc));
            }
            if  (TRUE == dbhasretstat(conn->dbproc)) {
                ifBLIN_QX1("Status returned %d", dbretstatus(conn->dbproc));
        }   }
        break;
    case PGOBLIN_CR_TUPL:
        if  (FAIL == dbfcmd(conn->dbproc, rct->text)) {
            ifBLIN_QX0("dbfcmd=%s~", rct->text);
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        if  (FAIL == dbsqlexec(conn->dbproc)) {
            ifBLIN_QX0("dbsqlexec=%s~", rct->text);
            ERROUT(EX_DATAERR, EDOOFUS);
        }
        for (int i = 0; NO_MORE_RESULTS != (erc = dbresults(conn->dbproc)); ++i) {
            int          ncol;

            ifBLIN_QX4("RESULT %d", i);
            if  (FAIL == erc) {
                ifBLIN_QX0("dbresults %d", i);
                ERROUT(EX_DATAERR, EDOOFUS);
            }
            if  (!!*rsv) clear(exenv, nou);
            if  (!(*rsv = resv(rdb))) {
                ifBLIN_QX0("no RAM");
                ERROUT(EX_OSERR, ENOMEM);
            }
            if  (0 > (collect(conn->dbproc, *rsv, 1))) {
                ifBLIN_QX0("collect");
                goto out;
            }
            if  (!(*rsv)->md) { /* XXXX */
                (*rsv)->md = mular_create(MULAR_FLAGS, 3, sizeof(char*), mdp);
                (*rsv)->md->tofree = free;
            }
            for (int j = 0; NO_MORE_ROWS != (row_code = dbnextrow(conn->dbproc)); ++j) {
                ifBLIN_QX4("ROW %d(%d)", j, row_code);
                switch (row_code) {
                case REG_ROW:
                    for (ncol = 0; ncol < (*rsv)->ncol; ++ncol) {
                        char **m;
                        char  *t;

                        ifBLIN_QX4("COL %d", ncol);
                        if  (0 < (*rsv)->status[ncol]) {
                            ifBLIN_QX0("%d x %d over %d", (*rsv)->nrow, ncol, (*rsv)->status[ncol]);
                            ERROUT(EX_SOFTWARE, EDOOFUS);
                        }
                        if  (!(m = mular_add((*rsv)->md))) {
                            ifBLIN_QX0("mular_add");
                            ERROUT(EX_SOFTWARE, EDOOFUS);
                        }
                        ifBLIN_QX7("m %"BLIN_X, BLIN_I(m));
                        if  (0 > (*rsv)->status[ncol]) {
                            ifBLIN_QX5("%d x %d NULL", (*rsv)->nrow, ncol);
                            *m = NULL;
                        } else {
                            ifBLIN_QX7( "[%"BLIN_D"]=%s~"
                                      , strlen((*rsv)->buffer[ncol])
                                      , (*rsv)->buffer[ncol]
                                      );
                            if  (!(t = strdup((*rsv)->buffer[ncol]))) {
                                ifBLIN_QX0("no RAM");
                                ERROUT(EX_OSERR, ENOMEM);
                            } else {
                                ifBLIN_QX7("t %"BLIN_X, BLIN_I(t));
                                if  (!!(*rsv)->reco) {
                                    for (char *u = t; !!*u; ++u) *u = (*rsv)->reco[*u & 0xFF];
                                }
                                ifBLIN_QX5("%d x %d =%s~", (*rsv)->nrow, ncol, t);
                                *m = t;
                    }   }   }
                    (*rsv)->nrow += 1;
                    break;
                case BUF_FULL:
                    ifBLIN_QX0("BUF_FULL *rsv=%d row=%d", i, j);
                    break;
                case FAIL:
                    ifBLIN_QX0("FAIL *rsv=%d row=%d", i, j);
                    ERROUT(-EX_SOFTWARE, EINVAL);
                default:
                    ifBLIN_QX0("%d ignored, *rsv=%d row=%d", ncol, i, j);
                    break;
            }   }
            for (ncol = 0; ncol < (*rsv)->ncol; ncol++) {
                free((*rsv)->buffer[ncol]);
                (*rsv)->buffer[ncol] = NULL;
            }
            free((*rsv)->buffer);
            (*rsv)->buffer = NULL;
            free((*rsv)->status);
            (*rsv)->status = NULL;
            if  (0 > DBCOUNT(conn->dbproc)) {
                ifBLIN_QX1("%d: row: DBCOUNT %d", i, DBCOUNT(conn->dbproc));
            } else if ((*rsv)->nrow == DBCOUNT(conn->dbproc)) {
            } else if (!(*rsv)->nrow) {
                ifBLIN_QX1("%d: row: 0 > DBCOUNT %d", i, DBCOUNT(conn->dbproc));
            } else if ((*rsv)->nrow == DBCOUNT(conn->dbproc)) {
                ifBLIN_QX0("%d: row: %d > DBCOUNT %d", i, (*rsv)->nrow, DBCOUNT(conn->dbproc));
            }
            if  (TRUE == dbhasretstat(conn->dbproc)) {
                ifBLIN_QX1("Status returned %d", dbretstatus(conn->dbproc));
        }   }
        break;
    case PGOBLIN_CR_BSEL:
    case PGOBLIN_CR_BTUPL:
    default:
        ifBLIN_QX0("illegal cond=%d", cond);
        ex = EX_UNAVAILABLE;
        errno = ENOTSUP;
    }
out:
    if  (!!conn) conn->res = *rsv;
    ifBLIN_QX2("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int64_t
/**********************************************************************
 **                                                                  **/
resinfo(pgoblin_exenv *exenv, pgoblin_nr nio, int kind) {           /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    pgoblin_realpq *rsv;
    pgoblin_rio    *rio;
    int64_t         ex = 0;

    GET_RIO(rio, exenv->options, nio);
    rsv = rio->pq;

    switch (kind) {
    case PGOBLIN_Ntuples:
        ex = rsv ? rsv->nrow : 0;
        break;
    case PGOBLIN_Nfields:
        ex = rsv ? rsv->ncol : 0;
        break;
    default             :
        ex = -EX_UNAVAILABLE;
        errno = ENOTSUP;
    }
out:
    return(ex);
#   undef blin_internal_flags
}

static const char *
/**************************************************************************
 **                                                                      **/
subinfo(pgoblin_exenv *exenv, pgoblin_nr nio, int kind, int64_t sub) {  /**
 **                                                                      **
 **************************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    pgoblin_rio  *rio;
    const char   *cx = NULL;
    int64_t       ex;

    GET_RIO(rio, exenv->options, nio);
    switch (kind) {
    case PGOBLIN_ColName:
        if  (!!rio->pq->colinfo && (rio->pq->ncol > sub)) cx = rio->pq->colinfo[sub].Name;
        break;
    case PGOBLIN_TypeNm:
        ex = valinfo(exenv, nio, PGOBLIN_TypeOid, 0, sub);
        if  ((TYPEMAX > ex) && (0 <= ex) && (!!types[ex][0])) cx = types[ex];
        break;
    default:;
    }
out:
    return(cx);
#   undef blin_internal_flags
}

static char *
/**********************************************************************
 **                                                                  **/
erinfo(pgoblin_exenv *exenv, pgoblin_nr nio) {                      /**
 **                                                                  **
 **********************************************************************/
    char *cx = NULL;

    return(cx);
}

static char *
/**********************************************************************
 **                                                                  **/
getvalue(pgoblin_exenv *exenv, pgoblin_nr nio, int row, int col) {  /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    pgoblin_rio  *rio;
    char         *c = NULL;

    GET_RIO(rio, exenv->options, nio);
    if  (!!rio->pq && !valinfo(exenv, nio, PGOBLIN_IsNull, row, col)) {
        if  (!!(rio->pq->flags & PGOBLIN_BUFF)) {
            c = rio->pq->buffer[INDX(row, col)];
            ifBLIN_QX4("buff=%s~", c);
        } else if (!!rio->pq->md) {
            c = *(char**)mular_getix(rio->pq->md, INDX(row, col));
            ifBLIN_QX4("mular=%s~", c);
        } else {
            ifBLIN_QX0("NOTHING instead of MULAR");
            errno = EDOOFUS;
    }   }
out:
    return(c);
#   undef blin_internal_flags
}

static int
/***********************************************************************************
 **                                                                               **/
getstream(pgoblin_exenv *exenv, pgoblin_nr ndb, pgoblin_nr nou, int quant) {     /**
 **                                                                               **
 ***********************************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    int              row_code;
    pgoblin_db_conn *conn;
    pgoblin_realpq **rsv;
    pgoblin_rdb     *rdb;
    pgoblin_rio     *rou;
    int              ex   = EX_OK;

    GET_CON(rdb, exenv->options, ndb);
    conn = rdb->conn;
    ifBLIN_QX2("+ %08X %d", conn->flags, quant);
    if  (!(conn->flags & PGOBLIN_STREAMI)) {
        ifBLIN_QX0("not COPYout state");
        ERROUT(-EX_DATAERR, ENODEV);
    }
    GET_RIO(rou, exenv->options, nou);
    if  (!(rsv = &rou->pq)) {
        ifBLIN_QX0("no res");
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (!*rsv) {
        conn->flags |= PGOBLIN_STREAME;
        goto out;
    }
    (*rsv)->flags &= ~PGOBLIN_BUFF;
    (*rsv)->nrow = 0;
    if  (NO_MORE_ROWS != (row_code = dbnextrow(conn->dbproc))) {
        int          ncol;
        char        *c;

        switch (row_code) {
        case REG_ROW:
            ifBLIN_QX3("row_code %d", row_code);
            (*rsv)->nrow = 1;
            (*rsv)->flags |= PGOBLIN_BUFF;
            for (ncol = 0; ncol < conn->res->ncol; ++ncol) {
                if  (!conn->res->status[ncol] && !!(c = conn->res->buffer[ncol])) {
                    for (int i = 0; !!c[i]; ++i) {
                        c[i] = conn->res->reco[c[i] & 0x0FF];
            }   }   }
            break;
        case BUF_FULL:
            ifBLIN_QX0("BUF_FULL");
            break;
        case FAIL:
            ifBLIN_QX0("FAIL");
            ERROUT(-EX_SOFTWARE, EINVAL);
        default:
            ifBLIN_QX0("ignored");
            break;
       }
    } else {
        if  (conn->flags & PGOBLIN_STREAMT) conn->flags |= PGOBLIN_STREAME;
        ex = 0;
    }
    if  (!(*rsv)->nrow) conn->flags |= PGOBLIN_STREAME;
out:
    ifBLIN_QX2("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
getcopy(pgoblin_exenv *exenv, pgoblin_nr ndb, void **buf) {         /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    int              row_code;
    pgoblin_db_conn *conn = NULL;
    pgoblin_rdb     *rdb;
    int              ex   = EX_OK;

    GET_CON(rdb, exenv->options, ndb);
    conn = rdb->conn;
    ifBLIN_QX2("+ %08X %08X %"BLIN_X, rdb->flags, conn->flags, buf);
    if  (!(conn->flags & PGOBLIN_STREAMI)) {
        ifBLIN_QX0("not COPYout state");
        ERROUT(-EX_DATAERR, ENODEV);
    }
    if  (!conn->res) {
        conn->flags |= PGOBLIN_STREAME;
        ifBLIN_QX3("res out %08X", conn->flags);
        goto out;
    }
    conn->res->nrow = 0;
    if  (NO_MORE_ROWS != (row_code = dbnextrow(conn->dbproc))) {
        int          ncol;
        char        *c;
        int          l;
        int          k;

        switch (row_code) {
        case REG_ROW:
            ifBLIN_QX3("row_code %d", row_code);
            for (ncol = 0, l = 2; ncol < conn->res->ncol; ++ncol) {
                if  (!conn->res->status[ncol]) {
                    l += strlen(conn->res->buffer[ncol]) * 2 + 1;
                } else {
                    l += 3;
                    if  (0 < conn->res->status[ncol]) {
                        ifBLIN_QX0( "col %d: real %d > buff %d"
                                  , ncol
                                  , conn->res->status[ncol]
                                  , conn->res->colinfo[ncol].MaxLength
                                  );
                        ERROUT(-EX_SOFTWARE, ENOMEM);
                }   }
                ifBLIN_QX5("[%d]l %d", ncol, l);
            }
            ifBLIN_QX4("Want malloc %d", l);
            if  (!(*buf = malloc(l))) {
                ifBLIN_QX0("no RAM");
                ERROUT(-EX_OSERR, ENOMEM);
            }
            k = 0;
            for (ncol = 0; ncol < conn->res->ncol; ++ncol) {
                ifBLIN_QO5 {
                    ifBLIN_QX5( "%d[%d]=%s~ %d[%d,%d]%d"
                              , conn->res->status[ncol]
                              , ncol
                              , dbcolname(conn->dbproc, ncol + 1)
                              , dbcoltype(conn->dbproc, ncol + 1)
                              , dbcollen(conn->dbproc, ncol + 1)
                              , dbprcollen(conn->dbproc, ncol + 1)
                              , dbdatlen(conn->dbproc, ncol + 1)
                              );
                    blin_dumb( 5
                             , dbdata(conn->dbproc, ncol + 1)
                             , dbdatlen(conn->dbproc, ncol + 1)
                             );
                }
                if  (!!conn->res->status[ncol] || !(c = conn->res->buffer[ncol])) {
                    (*(char**)buf)[k++]  = '\\';
                    (*(char**)buf)[k++]  = 'N';
                } else {
                    ifBLIN_QX3("col[%d]=%s~", ncol, c);
                    for (int i = 0; !!c[i]; ++i) {
                        switch(c[i]) {
                        case 0x08: (*(char**)buf)[k++] = '\\'; (*(char**)buf)[k++] = 'b' ; break;
                        case 0x09: (*(char**)buf)[k++] = '\\'; (*(char**)buf)[k++] = 't' ; break;
                        case 0x0A: (*(char**)buf)[k++] = '\\'; (*(char**)buf)[k++] = 'n' ; break;
                        case 0x0B: (*(char**)buf)[k++] = '\\'; (*(char**)buf)[k++] = 'v' ; break;
                        case 0x0C: (*(char**)buf)[k++] = '\\'; (*(char**)buf)[k++] = 'f' ; break;
                        case 0x0D: (*(char**)buf)[k++] = '\\'; (*(char**)buf)[k++] = 'r' ; break;
                        case '\\': (*(char**)buf)[k++] = '\\'; (*(char**)buf)[k++] = '\\'; break;
                        default:
                            if  (!conn->res->reco) {
                                (*(char**)buf)[k++] = c[i];
                            } else {
                                (*(char**)buf)[k++] = conn->res->reco[c[i] & 0x0FF];
                }   }   }   }
                (*(char**)buf)[k++] = (ncol < (conn->res->ncol - 1)) ? '\t' : '\n';
                (*(char**)buf)[k] = 0;
                ifBLIN_QX5("[%d]k %d", ncol, k);
            }
            conn->res->nrow = 1;
            ex = k;
            break;
        case BUF_FULL:
            ifBLIN_QX0("BUF_FULL");
            break;
        case FAIL:
            ifBLIN_QX0("FAIL");
            ERROUT(-EX_SOFTWARE, EINVAL);
        default:
            ifBLIN_QX0("ignored");
            break;
        }
        dbclrbuf(conn->dbproc, 999);
    } else {
        if  (conn->flags & PGOBLIN_STREAMT) conn->flags |= PGOBLIN_STREAME;
        ex = 0;
        ifBLIN_QX3("row end %08X", conn->flags);
    }
    if  (!conn->res->nrow) conn->flags |= PGOBLIN_STREAME;
out:
    ifBLIN_QX2("- %d %08X", ex, !conn ? 0 : conn->flags);
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
endstream(pgoblin_exenv *exenv, pgoblin_nr ndb) {                   /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    BLIN_flag        cond;
    pgoblin_db_conn *conn = NULL;
    pgoblin_rdb     *rdb;
    int              ex   = 0;

    GET_CON(rdb, exenv->options, ndb);
    conn = rdb->conn;
    switch((cond = rdb->flags & PGOBLIN_CR_MASK)) {
    case PGOBLIN_CR_SSEL:
    case PGOBLIN_CR_OUT:
        if  (!(conn->flags & PGOBLIN_STREAMI)) {
            ifBLIN_QX0("not stream state");
            ERROUT(-EX_DATAERR, ENODEV);
        }
        if  (0 > DBCOUNT(conn->dbproc)) {
            ifBLIN_QX1("DBCOUNT");
        } else {
            ifBLIN_QX3("row %d", DBCOUNT(conn->dbproc));
        }
        if  (TRUE == dbhasretstat(conn->dbproc)) {
            ifBLIN_QX1("Status returned %d", dbretstatus(conn->dbproc));
        }
        break;
    case PGOBLIN_CR_IN:
        if  (!(conn->flags & PGOBLIN_COPYIN)) {
            ifBLIN_QX0("not COPYin state");
            ERROUT(-EX_DATAERR, ENODEV);
        }
        if  (0 > DBCOUNT(conn->dbproc)) {
            ifBLIN_QX1("DBCOUNT");
        } else {
            ifBLIN_QX2("row %d", DBCOUNT(conn->dbproc));
        }
        if  (TRUE == dbhasretstat(conn->dbproc)) {
            ifBLIN_QX1("Status returned %d", dbretstatus(conn->dbproc));
        }
        break;
    default:
        ifBLIN_QX0("- %d", cond);
        ex = -EX_UNAVAILABLE;
        errno = ENOTSUP;
    }
out:
    if  (!!conn) conn->flags = 0;
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
finish(pgoblin_exenv *exenv, pgoblin_nr ndb) {                      /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    pgoblin_db_conn *conn;
    pgoblin_rdb     *rdb;
    int              ex = EX_OK;

    GET_CON(rdb, exenv->options, ndb);
    conn = rdb->conn;
    if  (!!conn) {
        if  (!!conn->dbproc) {
            dbclose(conn->dbproc);
            conn->dbproc = NULL;
        }
        free(rdb->conn);
        rdb->conn = NULL;
    }
out:
    return(ex);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
denote(pgoblin_exenv *exenv, pgoblin_nr ndb, void **note) {         /**
 **                                                                  **
 **********************************************************************/
    return(EX_OK);
}

pgoblin_dbases PGOBLIN_MODULE =
{ 0
, "3pGoblin-" VERS
, "mssql\0" VERS
, 0
, init
, fini
, shurecon
, query
, NULL     /* prepare  */
, NULL     /* execute  */
, resinfo
, subinfo
, erinfo
, valinfo
, getvalue
, getstream
, getcopy
, NULL     /* putcopy  */
, endstream
, NULL     /* tarry    */
, NULL     /* notifies */
, finish
, clear
, NULL     /* erconn   */
, NULL     /* prexec   */
, denote
, NULL     /* dummy1   */
, NULL     /* dummy2   */
};
